__author__ = 'kako'

from django.utils.functional import cached_property


class AggregatedGroup(object):
    """
    Generic object that constructs related objects like a Django model
    from queryset values() data.
    """
    def __init__(self, model, row_values):
        self._model = model
        self._row_values = row_values
        self._set_values()

    def __repr__(self):
        return u'<{} for {}>'.format(self.__class__.__name__,
                                     self._model.__name__)

    @cached_property
    def _data(self):
        """
        Cached data built from instance raw _values as a dictionary.
        """
        d = {}

        # Iterate all keys and values
        for k, v in self._row_values.items():
            # Split related model fields
            attrs = k.rsplit('__', 1)

            # Set value depending case
            if len(attrs) == 2:
                # Related model field, store nested
                fk, fn = attrs
                if fk not in d:
                    d[fk] = {}
                d[fk][fn] = v

            else:
                # Own model field, store directly
                d[k] = v

        # Return (+cache) data
        return d

    def _set_values(self):
        """
        Populate instance with given.
        """
        # Iterate all keys and values in data
        for k, v in self._data.items():
            # If it's a dict, process it (it's probably instance data)
            if isinstance(v, dict):
                try:
                    # Get related model from field (follow path)
                    rel_model = self._model
                    for attr in k.split('__'):
                        rel_model = getattr(rel_model, attr).field.related_model

                except AttributeError:
                    # Not a model, maybe it is a dict field (?)
                    pass

                else:
                    # It is a model, build instance from data
                    k = k.replace('__', '_')
                    v = rel_model(**v)

            # Set value
            setattr(self, k, v)



__author__ = 'kako'

from django.db.models.query import ValuesQuerySet
from django.db.models import ForeignKey

from .group import AggregatedGroup


class GroupByQuerySet(ValuesQuerySet):
    """
    Modified ValuesQuerySet that yields AggregatedGroup instances instead
    of dictionaries, which resemble the queryset's model in that all foreign
    related field values become actual model instances.
    """
    def iterator(self):
        # Same as in django
        extra_names = list(self.query.extra_select)
        field_names = self.field_names
        annotation_names = list(self.query.annotation_select)
        names = extra_names + field_names + annotation_names

        # Iterate results and yield AggregatedGroup instances
        for row in self.query.get_compiler(self.db).results_iter():
            data = dict(zip(names, row))
            obj = AggregatedGroup(self.model, data)
            yield obj


class GroupByMixin(object):
    """
    QuerySet mixin that adds a group_by() method, similar to values() but
    which returns AggregatedGroup instances when iterated instead of
    dictionaries.
    """
    @classmethod
    def _expand_group_by_fields(cls, model, fields):
        """
        Expand FK fields into all related object's fields to avoid future
        lookups.

        :param fields: fields to "group by"
        :return: expanded fields
        """
        # Containers for resulting fields and related model fields
        res = []
        related = {}

        # Add own fields and populate related fields
        for field_name in fields:
            if '__' in field_name:
                # Related model field: append to related model's fields
                fk_field_name, related_field = field_name.split('__', 1)
                if fk_field_name not in related:
                    related[fk_field_name] = [related_field]
                else:
                    related[fk_field_name].append(related_field)

            else:
                # Simple field, get the field instance
                model_field = model._meta.get_field(field_name)

                if isinstance(model_field, ForeignKey):
                    # It's a FK, get model
                    related_model = model_field.related_model

                    # Append all its fields with the correct prefix
                    res.extend('{}__{}'.format(field_name, f.column)
                               for f in related_model._meta.fields)

                else:
                    # It's a common field, just append it
                    res.append(field_name)

        # Resolve all related fields
        for fk_field_name, field_names in related.items():
            # Get field
            fk = model._meta.get_field(fk_field_name)

            # Get all fields for that related model
            related_fields = cls._expand_group_by_fields(fk.related_model,
                                                         field_names)

            # Append them with the correct prefix
            res.extend('{}__{}'.format(fk_field_name, f) for f in related_fields)

        # Return all fields
        return res

    def group_by(self, *fields):
        """
        Clone the queryset using GroupByQuerySet.

        :param fields:
        :return:
        """
        fields = self._expand_group_by_fields(self.model, fields)
        return self._clone(klass=GroupByQuerySet, setup=True, _fields=fields)



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""
__author__ = 'kako'

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()
setup(
    name='django_group_by',

    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.1.1',

    description='Group by arbitrary model fields',
    long_description=long_description,
    url='https://github.com/kako-nawao/django-group-by',
    author='Claudio Omar Melendrez Baeza',
    author_email='claudio.melendrez@gmail.com',
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Plugins',
        'Framework :: Django :: 1.8',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development',
    ],

    keywords='django grouping values models',
    packages=find_packages(exclude=['contrib', 'docs', 'test_app']),

    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'django==1.8'
    ],
    extras_require={},
    package_data={},

    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    data_files=[],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={},

    # Django test suite
    test_suite='runtests.run_tests'
)
